/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors:
 * 2009-
 *      Oscar Koeroo <okoeroo@nikhef.nl>
 *      Mischa Sall\'e <msalle@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *      Gerben Venekamp <venekamp@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 *  \file glexec_ipc.c
 *
 *  Provides functionality for ipc handling: payload and epilogue, signals
 *  handling etc.
 */

#include "glexec_ipc.h"
#include "glexec_exit.h"
#include "glexec_log.h"
#include "glexec_util.h"    /* for glexec_is_switching and umask */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/times.h>

/************************************************************************/
/* DEFINES AND TYPES                                                    */
/************************************************************************/

#define _PAYLOAD_STDINCLOSE	(1<<0)
#define _PAYLOAD_TAKETTY	(1<<1)

/************************************************************************/
/* GLOBAL VARIABLES                                                     */
/************************************************************************/

/* Stage of child process (eventual payload)
 * -3	child has ended
 * -2	child has not yet run
 * -1	fork failed
 * 0	process is child
 * >0	pid of child
 */
static pid_t global_ipc_childstage=-2;

/* payload child pid, -1 no child, 0 child itself. */
static pid_t global_ipc_childpid=-1;

/* Exit code of child, only valid when global_ipc_childstage==-3 */
static int global_ipc_childrc=0;

/* Stage of epilogue process
 * -4	epilogue won't run
 * -3	epilogue has ended
 * -2	epilogue has not yet run (but will)
 * -1	fork failed
 * 0	process is epilogue
 * >0	pid of epilogue
 */
static pid_t global_ipc_epilogstage=-4;

/* epilogue child pid, -1 no child, 0 child itself. */
static pid_t global_ipc_epilogpid=-1;

/* Exit code of epilogue, only valid when epiloguestage==-3 */
static int global_ipc_epilogrc=0;

/* Whether we can switch: this might change after e.g. lcmaps posix_enf */
static int global_ipc_canswitch=-1;

/* Number of TERM/KILL signals sent so far to the payload.
 * 0 nothing yet, 1 TERM, 2 KILL, 3 kill didn't succeed: premature exit */
static int global_ipc_numsigsent=0;

/* Grace period before which to test for surviving children */
static unsigned global_ipc_termdelay=GLEXEC_TERM_GRACE_TIME;
static unsigned global_ipc_killdelay=GLEXEC_KILL_GRACE_TIME;
static unsigned global_ipc_epilogtimeout=GLEXEC_EPILOG_TIMEOUT;

/* Whether gLExec should setpgid() on its children */
static int global_ipc_usesetpgid=1;

/* Whether gLExec should install handlers for SIGTRAP, SIGPOLL, SIGPROF etc.:
 * -1 unset, 1 set handlers for signals, 0 don't set handlers */
static int global_ipc_extrasighandlers=-1;

/* Whether gLExec will force the payload to run in the background, only relevant
 * when payload has different process group than gLExec: either payload or
 * gLExec will have the stdin tty. By default gLExec transfers to payload, this
 * flag keeps it at lingering gLExec and gives payload /dev/null */
static int global_ipc_forcepayloadbg=0;

/* UID of user calling gLExec: if root (s)he may still send signals in
 * post-payload stages */
static uid_t global_ipc_calleruid;

/* Options concerning payload stdin: whether it should take tty, or whether is
 * should close (reopen to /dev/null) stdin */
static unsigned int global_ipc_payloadstdinopts=0;

/* Whether lingering gLExec should close its stdin */
static int global_ipc_glexecclosestdin=0;

/* return value of waitpid when coming from an interrupt. */
static pid_t global_ipc_wpid=0;

/* status field of waitpid when coming from an interrupt. */
static int global_ipc_status=0;

/*************************************************************************/
/* FUNCTION PROTOTYPES                                                   */
/*************************************************************************/
/*!
 * Sends signal to pid (process group when negative)
 * \param signo signal to send
 * \param pid process ID or process group (when negative) to signal
 * \return 0 when signal was delivered, otherwise -1
 * When seteuid fails, the function might still return 0, if the signal can be
 * delivered
 */
int glexec_send_signal(pid_t pid, int signo, pid_t *stage);

/*!
 * Safely sends a signal to child pid, only when it's a child process
 * Temporarily blocks all signals (except KILL and STOP) between waitpid() and
 * kill() to prevent other signal handlers from calling us a second time.
 * waitpid() will loop until it's not interrupted to retrieve the wait status.
 * Signature of function is same as that of kill(). When there is no child,
 * errno will be ESRCH and return value -1.
 * \param pid process (group) to send signal
 * \param signo to send
 * \return 0 when successfully send signal, -1 on error
 */
int glexec_safe_kill(pid_t pid, int signo);

/*!
 * Handler for SIGINT, SIGTERM, SIGXCPU, SIGXFSZ: send children a TERM and set
 * alarm. */
void glexec_term_action(int, siginfo_t *, void *);

/*!
 * Handler for SIGALRM:
 * - glexec_term_action sets alarm for grace period after which to send a
 *   SIGKILL
 * - glexec_alarm_action sets alarm when a SIGKILL has been sent for a final
 *   graceperiod, after which it ends without further sending signals
 */
void glexec_alarm_action(int, siginfo_t *, void *);

/*!
 * Handler for SIGABRT and SIGSYS: typically not from kernel, but still called
 * when something is hopelessly wrong (e.g. failed malloc). sends a SIGTERM to
 * children, without setting alarm and exits.
 */
void glexec_abrt_action(int signo, siginfo_t *info, void *context);

/*!
 * Handler for SIGBUS, SIGFPE, SIGILL, SIGSEGV when received
 * from kernel: sends a SIGTERM to children, without setting alarm and exits.
 */
void glexec_core_action(int, siginfo_t *, void *);

/*!
 * Handler for ignore signals: log and ignore, e.g. SIGPIPE and SIGTSTP
 * (typically ctrl-z)
 */
void glexec_ignore_action(int signo, siginfo_t *info, void *context);

/*!
 * Changes effective and real identity (except secondary groups) to root to
 * prevent receiving of signals. This should be called either when receiving a
 * term-type signal, while there is a payload or when payload has finished.
 */
void glexec_continue_as_root(void);

/*!
 * Handler for other signals: forward to child (not children!)
 */
void glexec_other_action(int, siginfo_t *, void *);

/*!
 * Looks for processes in process group pid (when using process groups) and
 * logs if there are processes remaining.
 * \param pid
 */
void glexec_cleanup_procesgroup(pid_t pid);

/*!
 * Parse status from waitpid, log reason and return corresponding exitcode
 * \param status status obtained from waitpid()
 * \param pid process id for which status is parsed (only for logging purposes)
 * \return exit code corresponding to waitpid
 */
int glexec_parse_waitstatus(int status, pid_t pid);

/*!
 * Starts (started==0) and stops+logs (started==1) accounting info. When called
 * a second time with started==1, nothing happens.
 * \param started 0 for starting the timing records, 1 for stopping and
 * printing.
 */
void glexec_accounting(int started);

/*************************************************************************/
/* FUNCTION DECLARATIONS                                                 */
/*************************************************************************/

/*!
 * Sets grace time after sending TERM to child process, before sending KILL.
 * Default is GLEXEC_TERM_GRACE_TIME, 0 doesn't time-out
 * \param delay in seconds
 */
void glexec_set_term_delay(unsigned delay)  {
    global_ipc_termdelay=delay;
}

/*!
 * Sets grace time after sending KILL to child process, before exiting.
 * Default is GLEXEC_KILL_GRACE_TIME, 0 doesn't time-out (don't do that!)
 * \param delay in seconds
 */
void glexec_set_kill_delay(unsigned delay)  {
    global_ipc_killdelay=delay;
}

/*!
 * Sets caller id into global static.
 * \param global_ipc_calleruid;
 */
void glexec_set_caller_uid(uid_t glexec_global_ipc_calleruid)	{
    global_ipc_calleruid=glexec_global_ipc_calleruid;
}

/*!
 * Sets whether gLExec should call set handlers for signals used by debuggers
 * and the like, such as SIGTRAP.
 * \param value 1 to install custom handlers, 0 not to set them.
 */
void glexec_set_use_extra_sighandlers(int value)  {
    global_ipc_extrasighandlers=value;
}

/*
 * Sets whether gLExec will force the payload to run in the background, only
 * relevant when payload has different process group than gLExec: either payload
 * or gLExec will have the stdin tty. By default gLExec transfers to payload,
 * this flag keeps it at lingering gLExec and gives payload /dev/null */
void glexec_set_payload_should_run_in_background(int value) {
    global_ipc_forcepayloadbg=value;
}

/*!
 * Sets whether gLExec should call setpgid(0,0) in a child process and hence
 * send TERM and KILL signals to the process group or just the process id.
 * Default is to use setpgid()
 * \param value 1 to use setpgid 0 not to.
 */
void glexec_set_use_setpgid(int value)  {
    global_ipc_usesetpgid=value;
}

/*!
 * Sets grace time after start of epilogue script, before sending it a SIGTERM.
 * Default is GLEXEC_EPILOG_TIMEOUT, 0 doesn't time-out (not advisable)
 * \param delay in seconds
 */
void glexec_set_epilog_timeout(unsigned timeout)  {
    global_ipc_epilogtimeout=timeout;
}

/*!
 * Sets flag to make sure epilogue will run, if already called, or child is
 * already being forked, it will be ignored.
 */
void glexec_set_epilog_should_run(void)	{
    if (global_ipc_epilogstage==-4 && global_ipc_childstage==-2)
	global_ipc_epilogstage=-2;
}

/*!
 * Getter method for glexec_set_use_setpgid
 * \return 0 don't use process groups
 * \return 1 use process groups
 */
int glexec_get_use_setpgid(void)  {
    return global_ipc_usesetpgid;
}
/*!
 * Getter method for epilogue running...
 * \return 0 epilogue will not run
 * \return 1 other cases, including failed, finished etc.
 */
int glexec_get_epilog_should_run(void)    {
    return (global_ipc_epilogstage==-4 ? 0 : 1);
}

/*!
 * Getter method for child PID
 * \return pid_t of child process
 */
pid_t glexec_get_childpid(void)	{
    return global_ipc_childpid;
}

/*!
 * Getter method for child exit code
 * \return exit code of payload process
 */
int glexec_get_childrc(void)	{
    return global_ipc_childrc;
}

/*!
 * Getter method for epilogue PID
 * \return pid_t of child process
 */
pid_t glexec_get_epilogpid(void)	{
    return global_ipc_epilogpid;
}

/*!
 * Sends signal to pid (process group when negative)
 * \param pid process ID or process group (when negative) to signal
 * \param signo signal to send
 * \return 0 signal delivered
 * \return -1 signal failed
 * \return -2 invalid child pid, cannot kill
 */
int glexec_send_signal(pid_t pid, int signo, pid_t *stage)	{
    if (glexec_safe_kill(pid,signo)) { /* signal failed */
	if (errno==ESRCH)   {   /* invalid pid */
	    *stage=-3;  /* Assume child has finished */
	    alarm(0);	/* Make sure to cancel any running timers */
	    /* Log failure of kill() */
	    glexec_log(GLEXEC_LOG_NOTICE,
		"Cannot send signal %d to %d: %s.\n",signo,pid,strerror(errno));
	    /* Kill failed due to absense of child */
	    return -2;
	}
	/* Other reason: log failure of kill() */
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error: Cannot send signal %d to %d: %s.\n",
	    signo,pid,strerror(errno));
	/* Could not kill, child still existing */
	return -1;
    }
    /* All is fine */
    return 0;
}

/*!
 * Handler for SIGINT, SIGTERM, SIGXCPU, SIGXFSZ: send children a TERM and set
 * alarm. Behaviour depends on stage of payload and epilogue.
 */
void glexec_term_action(int signo, siginfo_t *info, void *context)   {
    /* Action depends on the stage were in */
    switch (global_ipc_childstage)	{
	case -2:    /* pre-child: log and exit */
	    /* Only accept process signals from caller or root */
	    if (info->si_code<=0 &&
		    info->si_uid!=global_ipc_calleruid && info->si_uid!=0)   {
		glexec_log(GLEXEC_LOG_NOTICE,
		    "Ignoring signal %d from pid=%d/uid=%u\n",
		    signo,info->si_pid,info->si_uid);
		return;
	    }
	    /* Origin is ok: log and exit */
	    glexec_log(GLEXEC_LOG_NOTICE,
		"Caught signal %d, exiting...\n", signo);
	    /* This will close necessary files etc. and then exit with a
	     * 128+signo */
	    glexec_exit(
		    (signo==SIGINT || signo==SIGTERM) ? 
			GLEXEC_SUCCESS :
			GLEXEC_UNCLASIFIED,
		    GLEXEC_EXECVE_SIGEXIT_BASE+signo);
	case -1:    /* fork failed: we received it during the exit: this should
		       not happen because glexec_exit blocks SIGTERM etc. */
	    glexec_log(GLEXEC_LOG_NOTICE,
		"Caught signal %d but fork failed so we are already exiting.\n",
		signo);
	    return;
	case 0:	    /* child itself before execve: log and exit */
	    glexec_log(GLEXEC_LOG_NOTICE,
		"Payload: caught signal %d, exiting...\n", signo);
	    exit(GLEXEC_EXECVE_SIGEXIT_BASE+signo);
	case -3:
	    /* post-child: action depends on epilogue: first handle
	     * non-authorized. On Linux a CTRL-c sends a non-process code
	     * (code>0) and we should check for global_ipc_calleruid, on MacOS
	     * it looks like a signal from a different process, si_code>=0. */
	    if ( (info->si_code<=0 && info->si_uid!=0) ||
		 (info->si_code>0  && global_ipc_calleruid !=0
				   && signo==SIGINT) ) {
		/* warn about CTRL-C also on stderr */
		if (signo==SIGINT)
		    fprintf(stderr,
			"[gLExec]: Ignoring caught SIGINT from non-root since "
			"payload has finished.\n");
		/* log message depends on type of signal */
		if (info->si_code<=0)
		    glexec_log(GLEXEC_LOG_NOTICE,
			"Ignoring signal %d from pid=%d/uid=%u, "
			"payload has already finished.\n",
			signo,info->si_pid,info->si_uid);
		else
		    glexec_log(GLEXEC_LOG_NOTICE,
			"Ignoring signal %d (si_code=%d), "
			"payload has already finished.\n",
			signo,info->si_code);
		return;
	    }
	    
	    /* Remainder is authorized signal: either kernel or root
	     * action depends on stage of epilogue. */
	    switch (global_ipc_epilogstage) {
		case -4: /* epilogue won't run */
		case -3: /* epilogue finished */
		case -1: /* epilogue fork failed */
		    /* Final shutdown, this should not take long so don't
		     * interrupt */
		    glexec_log(GLEXEC_LOG_NOTICE,
			"Ignoring signal %d during gLExec shutdown.\n",signo);
		    return;
		case -2:
		    /* epilogue hasn't run yet: accounting might not yet have
		     * been printed */ 
		    glexec_log(GLEXEC_LOG_WARNING,
			    "Warning: Caught signal %d before epilogue ran!\n",
			    signo);
		    glexec_accounting(1);
		    glexec_exit(GLEXEC_PREMATURE_EXIT,global_ipc_childrc);
		case 0:
		    /* epilogue itself, before execve() */
		    glexec_log(GLEXEC_LOG_WARNING,
			    "Epilogue: Caught signal %d before execve()'ed\n",
			    signo);
		    exit(GLEXEC_EXECVE_SIGEXIT_BASE+signo);
		default:
		    /* active epilogue is running: need to forward, but only
		     * forward something when we haven't yet sent a SIGTERM */
		    if (global_ipc_numsigsent>0)	{
			glexec_log(GLEXEC_LOG_NOTICE,
			    "Ignoring signal %d since epilogue group %d has "
			    "already been sent SIGTERM...\n",
			    signo,global_ipc_epilogpid);
			return;
		    }
		    /* Now will forward to epilogue, this is a warning as it
		     * will lead to a non-zero exit value of the epilogue and a
		     * hence a failure of gLExec. */
		    glexec_log(GLEXEC_LOG_WARNING,
			"Warning: Caught signal %d, "
			"sending SIGTERM to epilogue process %s %d\n",
			signo, global_ipc_usesetpgid ? "group" : "id",
			global_ipc_epilogpid);
		    /* warn about CTRL-C also on stderr */
		    if (signo==SIGINT)
			fprintf(stderr,
			    "[gLExec]: Caught SIGINT, "
			    "sending SIGTERM to epilogue.\n");
		    /* Send a SIGTERM to epilogue id/group */
		    switch (glexec_send_signal(
				global_ipc_usesetpgid ? -global_ipc_epilogpid :
							global_ipc_epilogpid,
				SIGTERM, &global_ipc_epilogstage))  {
			case 0:
			    /* Succesfully sent signal */
			    global_ipc_numsigsent=1;
			    alarm(global_ipc_termdelay);
			    return;
			case -1:
			    /* Signalling epilogue failed: exit */
			    glexec_exit(GLEXEC_PREMATURE_EXIT,
					global_ipc_childrc);
			case -2:
			default:
			    /* Epilogue has already finished: can ignore */
			    return;
		    }
		    /* Succesfully sent TERM, update counter, and set alarm */
	    }
	default:    /* global_ipc_childstage>0 i.e. there is a valid child */
	    /* Only forward something when we haven't yet sent a TERM */
	    if (global_ipc_numsigsent>0)	{
		/* SIGTERM has already been sent, we're in the
		 * global_ipc_termdelay grace period and waiting for normal
		 * child exit */
		glexec_log(GLEXEC_LOG_NOTICE,
		    "Caught signal %d but payload group %d has "
		    "already been sent SIGTERM...\n",signo,global_ipc_childpid);
		return;
	    }

	    /* Now become root to be immune for SIGKILL and SIGSTOP */
	    glexec_continue_as_root();

	    /* Now will forward to payload, this is a warning as it will lead to
	     * a non-zero exit value of the payload. */
	    glexec_log(GLEXEC_LOG_WARNING,
		"Warning: Caught signal %d, "
		"sending SIGTERM to payload process %s %d...\n",
		signo, global_ipc_usesetpgid ? "group" : "id",
		global_ipc_childpid);
	    /* Log CTRL-c on cmdline such that glexec user knows what's
	     * happening */
	    if (signo==SIGINT)
		fprintf(stderr,
		    "[gLExec]: Caught SIGINT, sending SIGTERM to payload.\n");
	    /* Send a SIGTERM to child */
	    switch (glexec_send_signal(
			global_ipc_usesetpgid ? -global_ipc_childpid :
						global_ipc_childpid,
			SIGTERM, &global_ipc_childstage))    {
		case 0:	/* Succesfully sent signal */
		    global_ipc_numsigsent=1;
		    alarm(global_ipc_termdelay);
		    return;
		case -1: /* Signalling failed: log accounting and exit */
		    glexec_accounting(1);
		    glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
		case -2:
		default: /* Child has already finished: can ignore */
		    return;
	    }
    }
}

/*!
 * Handler for SIGALRM:
 * - glexec_term_action sets alarm for grace period after which to send a
 *   SIGKILL (global_ipc_numsigsent==1)
 * - glexec_alarm_action sets alarm when a SIGKILL has been sent for a final
 *   graceperiod, after which it ends without further sending signals
 *   (global_ipc_numsigsent==2)
 * - startup of epilogue triggers a alarm after which it will be terminated
 *   (global_ipc_numsigsent==0, only for global_ipc_epilogstage>0 or ==-3)
 */
void glexec_alarm_action(int signo, siginfo_t *info, void *context)  {
    /* First handle different reasons to ignore the signal */

    /* Check the origin: should be from a timer (i.e. kernel) */
    if (info->si_code <= 0) {
	glexec_log(GLEXEC_LOG_NOTICE,
		"Ignoring ALARM signal (%d) since it is not sent by kernel "
		"(si_code=%d).\n",
		signo,info->si_code);
	return;
    }

    /* Handle active payload stage */
    if (global_ipc_childstage>0)	{
	switch (global_ipc_numsigsent) {
	    case 1: /* TERM has been sent */
		glexec_log(GLEXEC_LOG_WARNING,
		    "Grace time of %d sec expired, "
		    "sending SIGKILL to payload %s %d...\n",
		    global_ipc_termdelay,
		    global_ipc_usesetpgid ? "group" : "id",
		    global_ipc_childpid);
		switch (glexec_send_signal(
			    global_ipc_usesetpgid ? -global_ipc_childpid :
						    global_ipc_childpid,
			    SIGKILL, &global_ipc_childstage)) {
		    case 0:	/* Succesfully sent signal */
			global_ipc_numsigsent=2;
			alarm(global_ipc_killdelay);
			return;
		    case -1: /* Signalling failed: log accounting and exit */
			glexec_accounting(1);
			glexec_exit(GLEXEC_SIGNAL_ERROR,
				    GLEXEC_NO_CHILD_EXIT_CODE);
		    case -2:
		    default: /* Child has already finished: can ignore */
			return;
		}
	    case 2: /* KILL has been sent, but didn't lead to child ending after
		       global_ipc_killdelay: something is horribly broken  */
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: payload %d did not die from a SIGKILL )-:\n",
		    global_ipc_childpid);
		glexec_accounting(1);
		glexec_exit(GLEXEC_PREMATURE_EXIT,GLEXEC_NO_CHILD_EXIT_CODE);
	    default: /* unexpected */
		glexec_log(GLEXEC_LOG_WARNING,
		    "Warning: ignoring unexpected ALARM signal (%d) "
		    "(si_code=%d).\n", signo,info->si_code);
		return;
	}
    }

    /* Handle active epilogue stage */
    if (global_ipc_epilogstage>0)	{
	switch (global_ipc_numsigsent) {
	    case 0: /* total runtime has expired: send TERM */
		glexec_log(GLEXEC_LOG_WARNING,
		    "Maxmimum runtime of %d sec. expired for epilogue, "
		    "sending SIGTERM to its process %s %d...\n",
		    global_ipc_epilogtimeout,
		    global_ipc_usesetpgid ? "group" : "id",
		    global_ipc_epilogpid);
		switch (glexec_send_signal(
			    global_ipc_usesetpgid ? -global_ipc_epilogpid :
						    global_ipc_epilogpid,
			    SIGTERM, &global_ipc_epilogstage))  {
		    case 0: /* Succesfully sent signal */
			global_ipc_numsigsent=1;
			alarm(global_ipc_termdelay);
			return;
		    case -1: /* Signalling epilogue failed: exit */
			glexec_exit(GLEXEC_PREMATURE_EXIT,global_ipc_childrc);
		    case -2:
		    default: /* Epilogue has already finished: can ignore */
			return;
		}
		break;
	    case 1: /* TERM has been sent: send KILL */
		glexec_log(GLEXEC_LOG_WARNING,
		    "Grace time of %d sec expired for epilogue, "
		    "sending SIGKILL to its process %s %d...\n",
		    global_ipc_termdelay,
		    global_ipc_usesetpgid ? "group" : "id",
		    global_ipc_epilogpid);
		switch (glexec_send_signal(
			    global_ipc_usesetpgid ? -global_ipc_epilogpid :
						    global_ipc_epilogpid,
			    SIGKILL, &global_ipc_epilogstage)) {
		    case 0: /* Succesfully sent signal */
			global_ipc_numsigsent=2;
			alarm(global_ipc_killdelay);
			return;
		    case -1: /* Signalling epilogue failed: exit */
			glexec_exit(GLEXEC_PREMATURE_EXIT,global_ipc_childrc);
		    case -2:
		    default: /* Epilogue has already finished: can further
				ignore */
			return;
		}
		break;
	    case 2: /* KILL has been sent, but didn't lead to child ending
		       after global_ipc_killdelay: something is
		       horribly broken: exit */
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: epilogue %d did not die from a SIGKILL )-:\n",
		    global_ipc_epilogpid);
		glexec_exit(GLEXEC_PREMATURE_EXIT,global_ipc_childrc);
	    default: /* unexpected */
		glexec_log(GLEXEC_LOG_WARNING,
		    "Warning: ignoring unexpected ALARM signal (%d) "
		    "(si_code=%d).\n", signo,info->si_code);
		return;
	    break;
	}

	return;
    }

    /* Should be ignored: figure out the reason and log it */

    /* Epilogue already finished, but alarm still running */
    if (global_ipc_epilogstage==-3)    {
	glexec_log(GLEXEC_LOG_NOTICE,
		"Ignoring ALARM signal (%d) (si_code=%d): "
		"epilogue has already exited.\n",
		signo,info->si_code);
	return;
    }

    /* Child already finished, but alarm still running */
    if (global_ipc_childstage==-3 && 
	    (global_ipc_epilogstage==-4 || global_ipc_epilogstage==-2))	{
	glexec_log(GLEXEC_LOG_NOTICE,
		"Ignoring ALARM signal (%d) (si_code=%d): "
		"payload has already exited.\n",
		signo,info->si_code);
	return;
    }

    /* Any other reason is unexpected and should not happen */
    glexec_log(GLEXEC_LOG_WARNING,
	    "Warning: ignoring unexpected ALARM signal (%d) (si_code=%d).\n",
	    signo,info->si_code);
    return;
}

/*!
 * Handler for SIGABRT: typically not from kernel, but still called when
 * something is hopelessly wrong (e.g. failed malloc). Also use for SIGSYS.
 * sends a SIGTERM to children, without setting alarm and exits.
 */
void glexec_abrt_action(int signo, siginfo_t *info, void *context)  {
    /* Check the signal is not send from another process */
    if (info->si_code<=0 && info->si_pid!=getpid()) {
	glexec_log(GLEXEC_LOG_NOTICE,
		"Ignoring signal %d from pid=%d/uid=%ld "
		"(only acceptable from kernel).\n",
		signo,info->si_pid,info->si_uid);
	return;
    }

    /* We're in trouble: force logging to syslog! */
    glexec_set_log_to_syslog();

    /* Handle according to stage */
    if (global_ipc_childstage>0)	{ /* Active child: terminate */
	glexec_log(GLEXEC_LOG_ERROR,
		"GLExec received signal %d, this looks like a BUG, "
		"sending TERM to payload process %s %d and exiting.\n",
		signo,global_ipc_usesetpgid ? "group" : "id",
		global_ipc_childpid);
	glexec_send_signal(
		global_ipc_usesetpgid ? -global_ipc_childpid :
					global_ipc_childpid,
		SIGTERM, &global_ipc_childstage);
	/* Log accounting information */
	glexec_accounting(1);
    } else if (global_ipc_epilogstage>0)	{ /* Active epilog: terminate */
	glexec_log(GLEXEC_LOG_ERROR,
		"GLExec received signal %d, this looks like a BUG, "
		"sending TERM to epilogue process %s %d and exiting.\n",
		signo,global_ipc_usesetpgid ? "group" : "id",
		global_ipc_epilogpid);
	glexec_send_signal(
		global_ipc_usesetpgid ? -global_ipc_epilogpid : 
					global_ipc_epilogpid,
		SIGTERM, &global_ipc_epilogstage);
    } else  {
	glexec_log(GLEXEC_LOG_ERROR,
		"GLExec received signal %d, this looks like a BUG.\n",
		signo);
    }

    /* When child has finished log its exit code since we'll lose it */
    if (global_ipc_childstage==-3)
	glexec_log(global_ipc_childrc ? GLEXEC_LOG_ERROR : GLEXEC_LOG_NOTICE,
		"Payload %d exited with exit code %d\n",
		global_ipc_childpid,global_ipc_childrc);
		
    /* Exit with a 128+signal */
    exit(GLEXEC_EXECVE_SIGEXIT_BASE+signo);
}

/*!
 * Handler for SIGBUS, SIGFPE, SIGILL, SIGSEGV when received
 * from kernel: sends a SIGTERM to children, without setting alarm and exits.
 */
void glexec_core_action(int signo, siginfo_t *info, void *context)  {
    /* Check the signal is not send from another process */
    if (info->si_code<=0) {
	glexec_log(GLEXEC_LOG_NOTICE,
		"Ignoring signal %d from pid=%d/uid=%ld "
		"(only acceptable from kernel).\n",
		signo,info->si_pid,info->si_uid);
	return;
    }

    /* We're in trouble: force logging to syslog! */
    glexec_set_log_to_syslog();

    /* Handle according to stage */
    if (global_ipc_childstage>0)	{ /* Active child: terminate */
	glexec_log(GLEXEC_LOG_ERROR,
		"GLExec received signal %d, address 0x%x, "
		"this looks like a BUG, "
		"sending TERM to payload process %s %d and exiting.\n",
		signo,info->si_addr, global_ipc_usesetpgid ? "group" : "id",
		global_ipc_childpid);
	glexec_send_signal(
		global_ipc_usesetpgid ? -global_ipc_childpid :
					global_ipc_childpid,
		SIGTERM, &global_ipc_childstage);
	/* Log accounting information */
	glexec_accounting(1);
    } else if (global_ipc_epilogstage>0)	{ /* Active epilog: terminate */
	glexec_log(GLEXEC_LOG_ERROR,
		"GLExec received signal %d, address 0x%x, "
		"this looks like a BUG, "
		"sending TERM to epilogue process %s %d and exiting.\n",
		signo,info->si_addr, global_ipc_usesetpgid ? "group" : "id",
		global_ipc_epilogpid);
	glexec_send_signal(
		global_ipc_usesetpgid ? -global_ipc_epilogpid :
					global_ipc_epilogpid,
		SIGTERM, &global_ipc_epilogstage);
    } else  {
	glexec_log(GLEXEC_LOG_ERROR,
		"GLExec received signal %d, address 0x%x, "
		"this looks like a BUG.\n",
		signo,info->si_addr);
    }

    /* When child has finished log its exit code since we'll lose it */
    if (global_ipc_childstage==-3)
	glexec_log(global_ipc_childrc ? GLEXEC_LOG_ERROR : GLEXEC_LOG_NOTICE,
		"Payload %d exited with exit code %d\n",
		global_ipc_childpid,global_ipc_childrc);
		
    /* Exit with a 128+signal */
    exit(GLEXEC_EXECVE_SIGEXIT_BASE+signo);
}

/*!
 * Handler for ignore signals: log and ignore, e.g. SIGPIPE
 */
void glexec_ignore_action(int signo, siginfo_t *info, void *context)  {
    glexec_log(GLEXEC_LOG_NOTICE,
	    "Ignoring received signal %d (si_code=%d)\n",
	    signo,info->si_code);
}

/*!
 * Handler for other signals: forward to child (not children!)
 */
void glexec_other_action(int signo, siginfo_t *info, void *context)  {
    /* Only forward the signal to payload when there is one, it hasn't been
     * send a TERM yet. Also forward when it's not from a process, it might be
     * CTRL-Z for example. */
    if (global_ipc_childstage>0 && global_ipc_numsigsent==0)  {
	/* Close stdin when receiving a SIGHUP (e.g. close shell) */
	if (signo==SIGHUP &&
	    ( (info->si_code<=0 && info->si_uid==global_ipc_calleruid) ||
	      (info->si_code>0) ) && 
	    global_ipc_glexecclosestdin==0)
	{
	    glexec_log(GLEXEC_LOG_NOTICE,
		"Caught SIGHUP, redirecting stdin for lingering gLExec "
		"to /dev/null.\n");
	    /* Reopen /dev/null as stdin, failure is NOT considered fatal. */
	    if (freopen("/dev/null","r",stdin)==NULL)
		glexec_log(GLEXEC_LOG_WARNING,
		    "Redirecting stdin for lingering gLExec failed: %s\n",
		    strerror(errno));
	    else    {
		global_ipc_glexecclosestdin=1;
		global_ipc_payloadstdinopts&=~_PAYLOAD_TAKETTY;
	    }
	}

	/* Log ctrl-z on cmdline such that user knows what's happening */
	if (signo==SIGTSTP)
	    fprintf(stderr,"[gLExec]: Forwarding caught SIGTSTP to payload.\n");
	
	/* Active child: forward */
	if (info->si_code<=0)
	    glexec_log(GLEXEC_LOG_NOTICE,
		    "Forwarding signal %d from pid=%d/uid=%ld to pid %d.\n",
		    signo,info->si_pid,info->si_uid,global_ipc_childpid);
	else
	    glexec_log(GLEXEC_LOG_NOTICE,
		    "Forwarding signal %d (si_code=%d) to pid %d.\n",
		    signo,info->si_code,global_ipc_childpid);

	/* Send signo (not only SIGTSTP) to payload */
	glexec_send_signal(global_ipc_childpid,signo,&global_ipc_childstage);

    } else  {
	/* Only special action when CTRL-Z is received in running
	 * global_ipc_epilogstage */
	if (global_ipc_epilogstage>0 && signo==SIGTSTP)	{
	    /* when we were called by root (Linux) or signalled by root (MacOS)
	     * forward the signal to payload */
	    if ( (info->si_code>0 && global_ipc_calleruid==0) ||
		 (info->si_code<=0 && info->si_uid==0) )    {
		if ( info->si_code<=0)
		    glexec_log(GLEXEC_LOG_NOTICE,
			"Forwarding signal %d from pid=%d/uid=%u to pid %d.\n",
			signo,info->si_pid,info->si_uid,global_ipc_epilogpid);
		else
		    glexec_log(GLEXEC_LOG_NOTICE,
			"Forwarding signal %d (si_code=%d) to pid %d.\n",
			signo,info->si_code,global_ipc_epilogpid);
		/* Log CTRL-z on cmdline such that glexec user knows what's
		 * happening */
		fprintf(stderr,
		    "[gLExec]: Forwarding caught SIGTSTP to epilogue.\n");
		glexec_send_signal(global_ipc_epilogpid,signo,
				   &global_ipc_epilogstage);
	    } else {
		/* Signal comes from either a non-root glexec caller (e.g.
		 * Linux) or non-root process (probably ctrl-z on MacOS) */
		if ( info->si_code<=0)
		    glexec_log(GLEXEC_LOG_NOTICE,
			"Ignoring signal SIGTSTP from pid=%d/uid=%u, "
			"payload has already finished.\n",
			signo,info->si_pid,info->si_uid);
		else
		    glexec_log(GLEXEC_LOG_NOTICE,
			"Ignoring signal SIGTSPT: "
			"payload has already finished.\n",signo);
		/* Log on cmdline such that user knows what's happening */
		fprintf(stderr,
		    "[gLExec]: Ignoring caught SIGTSTP from non-root since "
		    "payload has finished.\n");
	    }
	} else
	    /* all other cases should be ignored */
	    glexec_log(GLEXEC_LOG_NOTICE,
		    "Ignoring received signal %d (si_code=%d)\n",
		    signo,info->si_code);
    }
}

/*!
 * Installs signal handlers
 */
void glexec_install_sig_handlers(void)	{
    struct sigaction act;

    glexec_log(GLEXEC_LOG_DEBUG,"Installing signal handlers\n");

    /* Make sure there are no previous alarms */
    alarm(0);

    /* When not yet set, set the global_ipc_canswitch flag. We are normally
     * pre-lcmaps here, so switching mode is reliable. */
    if (global_ipc_canswitch==-1)
	global_ipc_canswitch=glexec_is_switching();

    /* Setup general blocking mask: all but non-blockable ones */
    if (sigfillset(&act.sa_mask) || 
	sigdelset(&act.sa_mask,SIGKILL) ||
	sigdelset(&act.sa_mask,SIGABRT) ||
	sigdelset(&act.sa_mask,SIGSTOP))    {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }
    act.sa_flags=SA_SIGINFO | SA_RESTART; /* Make sure to restart after
					     interrupt for e.g. waitpid() */

    /* Setup different handlers:
     * - Start with segv type handlers
     * - Remove the core type from the mask set and install the term handlers
     * - install the handlers only needed once we have a payload
     */

    /* ABORT type: SIGABRT, SIGSYS */
    act.sa_sigaction=glexec_abrt_action;
    if (sigaction(SIGABRT,&act,NULL) ||
	sigaction(SIGSYS,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* CORE type: SIGBUS, SIGFPE, SIGILL, SIGSEGV and if present SIGSTKFLT */
    act.sa_sigaction=glexec_core_action;
    if (sigaction(SIGBUS,&act,NULL) ||
	sigaction(SIGFPE,&act,NULL) ||
#ifdef SIGSTKFLT
	sigaction(SIGSTKFLT,&act,NULL) ||
#endif
	sigaction(SIGILL,&act,NULL) ||
	sigaction(SIGSEGV,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* TERM type: SIGINT, SIGTERM, SIGXCPU, SIGXFSZ. */
    act.sa_sigaction=glexec_term_action;
    if (sigdelset(&act.sa_mask,SIGBUS) ||
	sigdelset(&act.sa_mask,SIGFPE) ||
	sigdelset(&act.sa_mask,SIGILL) ||
	sigdelset(&act.sa_mask,SIGSEGV) ||
	sigdelset(&act.sa_mask,SIGSYS) ||
	sigaction(SIGINT,&act,NULL) ||
	sigaction(SIGTERM,&act,NULL) ||
	sigaction(SIGXCPU,&act,NULL) ||
	sigaction(SIGXFSZ,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* ALARM type: SIGALRM */
    act.sa_sigaction=glexec_alarm_action;
    if (sigaction(SIGALRM,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* IGNORE type */
    act.sa_sigaction=glexec_ignore_action;
    if (sigaction(SIGPIPE,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* OTHER type: include SIGTSTP (ctrl-z in terminal) and SIGCONT, also
     * I/O related ones such as SIGPOLL and SIGURG */
    act.sa_sigaction=glexec_other_action;
    if (sigaction(SIGHUP,&act,NULL) ||
	sigaction(SIGQUIT,&act,NULL) ||
	sigaction(SIGTSTP,&act,NULL) ||
	sigaction(SIGCONT,&act,NULL) ||
	sigaction(SIGURG,&act,NULL) ||
#ifdef SIGPOLL
	/* Note: SIGPOLL though POSIX is synonym for SIGIO on Linux and does not
	 * exist on MacOS */
	sigaction(SIGPOLL,&act,NULL) ||
#endif
#ifdef SIGIO
	/* Note: SIGPOLL and SIGIO on Linux are the same */
	sigaction(SIGIO,&act,NULL) ||
#endif
#ifdef SIGPWR
	/* Set a handler for SIGPWR when available, since it's it could lead to
	 * terminate on Linux: log and forward when possible */
	sigaction(SIGPWR,&act,NULL) ||
#endif
#ifdef SIGWINCH
	/* Set a handler for SIGWINCH when available, since it's it might be
	 * useful for the payload: log and forward when possible */
	sigaction(SIGWINCH,&act,NULL) ||
#endif
	sigaction(SIGUSR1,&act,NULL) ||
	sigaction(SIGUSR2,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }
    
    /* Make sure to really ignore SIGTTIN and SIGTTOU or we can't do jobcontrol
     * on the payload */
    act.sa_flags=0;
    act.sa_handler=SIG_IGN;
    if (sigaction(SIGTTIN,&act,NULL) ||
	sigaction(SIGTTOU,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }

}

/*!
 * Install optional extra signal handlers. We only can install these after
 * having set the flag global_ipc_extrasighandlers using the
 * glexec_set_use_extra_sighandlers setter function.
 */
void glexec_install_extra_sig_handlers(void) {
    struct sigaction act;
#if defined(SIGRTMIN) && defined(SIGRTMAX)
    int signo;
#endif

    if (global_ipc_extrasighandlers!=1)
	return;

    /* Setup general blocking mask: all but non-blockable ones */
    if (sigfillset(&act.sa_mask) || 
	sigdelset(&act.sa_mask,SIGKILL) ||
	sigdelset(&act.sa_mask,SIGABRT) ||
	sigdelset(&act.sa_mask,SIGSTOP))    {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing extra signal handlers: %s\n",
	    strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }
    act.sa_flags=SA_SIGINFO | SA_RESTART; /* Make sure to restart after
					     interrupt for e.g. waitpid() */
    /* OTHER type: include SIGTSTP (ctrl-z in terminal) and SIGCONT, also
     * I/O related ones such as SIGPOLL and SIGURG */
    act.sa_sigaction=glexec_other_action;

    /* Debug type handlers: SIGTRAP etc. These might prevents the use of a
     * debugger: log and forward when possible */
#ifdef SIGEMT
    glexec_log(GLEXEC_LOG_DEBUG,
	"Installing signal handlers for SIGTRAP, SIGEMT, SIGVTALRM, SIGPROF\n");
#else
    glexec_log(GLEXEC_LOG_DEBUG,
	"Installing signal handlers for SIGTRAP, SIGVTALRM, SIGPROF\n");
#endif
    if (sigaction(SIGTRAP,&act,NULL) ||
#ifdef SIGEMT
	sigaction(SIGEMT,&act,NULL) ||
#endif
	sigaction(SIGVTALRM,&act,NULL) ||
	sigaction(SIGPROF,&act,NULL))   {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",
	    strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Real-time signal handlers. These might interfere with profilers and tools
     * such as valgrind */
#if defined(SIGRTMIN) && defined(SIGRTMAX)
    glexec_log(GLEXEC_LOG_DEBUG,
	"Installing realtime signal handlers (SIGRTMIN=%d, SIGRTMAX=%d).\n",
	SIGRTMIN,SIGRTMAX);
    for (signo=SIGRTMIN; signo<=SIGRTMAX; signo++)	{
	/* Install Real-time signal handler, but don't exit on failure */
	if (sigaction(signo,&act,NULL))
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error while initializing real-time sighandler for %d: %s\n",
		signo,strerror(errno));
    }
#endif
}

/*!
 * Blocks all signals except for:
 * unblockable:
 * SIGKILL, SIGABRT, SIGSTOP
 * core-type signals: SIGBUS, SIGFPE, SIGILL, SIGSEGV, SIGSYS
 * \param oldsigset previous signal mask
 */
void glexec_block_signals(sigset_t *old_blocked_signals_set)	{
    static sigset_t blocked_signals_set;
    static int blocked_init=0;

    if (!blocked_init)	{ /* Whether blocked set is initialized */
	/* Setup blocked set */
    if (sigfillset(&blocked_signals_set) ||
	sigdelset(&blocked_signals_set,SIGKILL) ||
	sigdelset(&blocked_signals_set,SIGABRT) ||
	sigdelset(&blocked_signals_set,SIGSTOP) ||
	sigdelset(&blocked_signals_set,SIGBUS) ||
	sigdelset(&blocked_signals_set,SIGFPE) ||
	sigdelset(&blocked_signals_set,SIGILL) ||
	sigdelset(&blocked_signals_set,SIGSEGV) ||
	sigdelset(&blocked_signals_set,SIGSYS))	{
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error: cannot setup blocked signal set: %s",
		strerror(errno));
	    glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
	}
	blocked_init=1;
    }
    /* Now block the set and save the old set */
    if (sigprocmask(SIG_SETMASK,&blocked_signals_set,old_blocked_signals_set)) {
	glexec_log(GLEXEC_LOG_ERROR,
		"Error: cannot block signal set: %s", strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);;
    }
}

/*!
 * Set blocked signal mask to sigset, exit when fails
 * \param sigset signal set to block
 */
void glexec_sigprocmask(const sigset_t *masksigset) {
    if (sigprocmask(SIG_SETMASK,masksigset,NULL))  {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error: cannot set signal mask: %s", strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);;
    }
}

/*!
 * Getter whether glexec_ipc thinks glexec can switch. This is only reliable
 * after glexec_install_sig_handlers() has been called
 * \return 1 when glexec can switch 0 otherwise */
int glexec_can_switch(void) {
    return global_ipc_canswitch;
}

/*!
 * Changes effective and real identity (except secondary groups) to root to
 * prevent receiving of signals. This should be called either when receiving a
 * term-type signal, while there is a payload or when payload has finished.
 */
void glexec_continue_as_root(void)	{
    if (global_ipc_canswitch)	{
	if (seteuid(0) || setgid(0) || setreuid(0,0))  {
	    /* Update the global_ipc_canswitch flag, aparently we lost it's
	     * capability */
	    global_ipc_canswitch=0;
	    glexec_log(GLEXEC_LOG_WARNING,
		"Warning: Failed to setreuid(0,0) or setgid(0) "
		"will continue to linger as uid/euid=%d/%d: %s\n",
		getuid(),geteuid(),strerror(errno));
	} else 
	    glexec_log(GLEXEC_LOG_DEBUG,"gLExec now lingers as uid/euid=0/0\n");
    }
}

/*!
 * Looks for processes in process group pid (when using process groups) and
 * logs if there are processes remaining.
 * \param pid
 */
void glexec_cleanup_procesgroup(pid_t pid)    {
    /* If we're using process groups, we can still send a signal to the process
     * group in case there are some daemonized processes left. */
    if (global_ipc_usesetpgid)	{
	/* If pid is a pgid, test if there are grandchild processes in the same
	 * pgid, to be able to log the information */
	if (kill(-pid,0)==0)	{
	    glexec_log(GLEXEC_LOG_WARNING,
		"Warning: Child %d finished, but there are processes in "
		"group %d, probably daemonized grandchildren\n",
		pid,pid);
	}
	else if (errno!=ESRCH)
	    glexec_log(GLEXEC_LOG_NOTICE,
		"Could not get status for processgroup %d while looking for "
		"daemonized grandchildren: %s.\n",
		pid,strerror(errno));
    }
}

/*!
 * Safely sends a signal to child pid, only when it's a child process
 * Temporarily blocks all signals (except KILL and STOP) between waitpid() and
 * kill() to prevent other signal handlers from calling us a second time.
 * waitpid() will loop until it's not interrupted to retrieve the wait status.
 * Signature of function is same as that of kill(). When there is no child,
 * errno will be ESRCH and return value -1.
 * \param pid process (group) to send signal
 * \param signo to send
 * \return 0 when successfully send signal, -1 on error
 */
int glexec_safe_kill(pid_t pid, int signo)	{
    uid_t euid;
    int rc,rc_seteuid1=0,rc_seteuid2=0;
    int werrno,kerrno;
    sigset_t blockedset,oldset;

    /* Block all signals, including critical ones, make sure not to do anything
     * except most simple code between, no mallocs etc. */
    if (sigfillset(&blockedset) ||
	sigprocmask(SIG_SETMASK,&blockedset,&oldset))	{
	/* Not really possible, no errors are defined for the above calls */
	glexec_log(GLEXEC_LOG_ERROR,"Cannot block signals: %s\n",
		    strerror(errno));
	return -1;
    }

    /* Loop over waitpid in case it gets interrupted */
    do {
	if (global_ipc_wpid>0)	{
	    /* We have already retrieved status: no child */
	    if (sigprocmask(SIG_SETMASK,&oldset,NULL))
		glexec_log(GLEXEC_LOG_ERROR,"Error in sigprocmask(): %s\n",
			    strerror(errno));
	    errno=ESRCH;
	    return -1;
	}
        global_ipc_wpid=waitpid(pid, &global_ipc_status, WNOHANG);
    } while (global_ipc_wpid==-1 && errno==EINTR);
    werrno=errno;
    
    /* waitpid returns >0 when finished, ==0 when running, <0 on error */
    if (global_ipc_wpid==0) { /* Active child: forward signal */
	/* do geteuid here instead of inside following if block to prevent
	 * compiler warning */
	euid=geteuid();
	/* Switch to effective root if possible, don't log anything now */
	if (global_ipc_canswitch)
	    if ((rc_seteuid1=seteuid(0))!=0) global_ipc_canswitch=0;
	rc=kill(pid,signo); /* errno now of kill */
	kerrno=errno;	/* save errno */
	if (global_ipc_canswitch)
	    if ((rc_seteuid2=seteuid(euid))!=0) global_ipc_canswitch=0;

	/* Reset procmask */
	if (sigprocmask(SIG_SETMASK,&oldset,NULL))
	    glexec_log(GLEXEC_LOG_ERROR,"Error in sigprocmask(): %s\n",
			strerror(errno));

	/* Log if userswitching went wrong */
	if (rc_seteuid1 || rc_seteuid2)
	    glexec_log(GLEXEC_LOG_ERROR,
		"This should not happen: could not userswitch, "
		"now lingering as uid/euid=%d/%d\n",getuid(),geteuid());

	/* Reset errno to that of kill */
	errno=kerrno;
	return rc;
    }

    /* Reset procmask */
    if (sigprocmask(SIG_SETMASK,&oldset,NULL))  /* errno is of sigprocmask */
	glexec_log(GLEXEC_LOG_ERROR,"Error in sigprocmask(): %s\n",
		    strerror(errno));

    /* No active child */
    if (global_ipc_wpid>0 || werrno==ECHILD)  {
	errno=ESRCH;
	return -1;
    }
   
    /* Something went wrong in waitpid */
    glexec_log(GLEXEC_LOG_ERROR,"Error in waitpid() for pid %d: %s\n",
		pid,strerror(werrno));

    /* Reset errno to that of waitpid */
    errno=werrno;
    return -1;
}

/*!
 * Parse status from waitpid, log reason and return corresponding exitcode
 * \param status status obtained from waitpid()
 * \param pid process id for which status is parsed (only for logging purposes)
 * \return exit code corresponding to waitpid
 */
int glexec_parse_waitstatus(int status, pid_t pid)	{
    int sig,rc;

    /* Check the exit code of child process */
    if (WIFEXITED(status))  { /* Normal program end: use WEXITSTATUS() */
	/* Note: WEXITSTATUS only when WIFEXITED is true */
	rc=WEXITSTATUS(status);
    } else {	/* Abnormal program end */
	if (WIFSIGNALED(status))    {/* Signal ended program: use WTERMSIG() */
	    /* Bash convention: http://tldp.org/LDP/abs/html/exitcodes.html */
	    sig=WTERMSIG(status);
	    rc=GLEXEC_EXECVE_SIGEXIT_BASE+sig;
	    glexec_log(GLEXEC_LOG_ERROR,
		"Child process %d exited with return code %d "
		"due to signal %d.\n", pid, rc, sig);
	} else {    /* Unknown reason for termination */
	    rc=1; /* No really reasonable alternative... */
	    glexec_log(GLEXEC_LOG_ERROR,
		"Child process %d exited abnormally with return code %d "
		"(status=%d).\n", pid, rc, status);
	}
	return rc;
    }
	
    /* Log exit status */
    glexec_log(rc ? GLEXEC_LOG_ERROR : GLEXEC_LOG_INFO,
	"Child process %d exited with return code %d.\n", pid, rc);

    return rc;
}

/*!
 * Waits for the child/payload process (in linger mode) to finish and correctly
 * handles, logs etc. the reason (normal exit, signal etc.). Timing statistics
 * are also logged.
 * \return exit code for child process, following (ba)sh convention: when due to
 * a signal: 128+<signal>, in case of unsignalled but abnormal program end, we
 * return 1. See http://tldp.org/LDP/abs/html/exitcodes.html for details.
 */ 
int glexec_waitchild(void)  {
    pid_t wpid,glexec_pgid=getpgrp();
    int status,wait_errno;

    /* Log at least to debug what we are doing */
    glexec_log(GLEXEC_LOG_DEBUG,"Waiting for payload process %d\n",
		global_ipc_childpid);

    /* Wait for pid to end, note: this is the only wait(pid), and the
     * sighandlers restart waitpid (SA_RESTART). Also wait for stopped and
     * continued processes to give feedback. */
    do	{
	wpid = waitpid(global_ipc_childpid, &status, WCONTINUED|WUNTRACED);
	wait_errno=errno;
	/* Test for stopped or continued payload */
	if (wpid>0) {
	    if (WIFSTOPPED(status)) { /* stopped payload */
		glexec_log(GLEXEC_LOG_NOTICE,
			"Payload %d has suspended.\n",(int)wpid);
		fprintf(stderr, "[gLExec]: Payload %d suspended.\n",(int)wpid);
		/* Take controlling terminal if needed */
		if ( global_ipc_payloadstdinopts & _PAYLOAD_TAKETTY) {
		    if (tcsetpgrp(STDIN_FILENO, glexec_pgid))   {
			global_ipc_payloadstdinopts&=~_PAYLOAD_TAKETTY;
			glexec_log(GLEXEC_LOG_WARNING,
			    "Cannot grab control of tty from payload: %s\n",
			    strerror(errno));
		    } else  
			glexec_log(GLEXEC_LOG_DEBUG,
			    "Payload suspended, taking back tty.\n");
		}
		continue;
	    } else if (WIFCONTINUED(status))	{ /* continued payload */
		glexec_log(GLEXEC_LOG_NOTICE,
			"Payload %d has continued.\n",(int)wpid);
		fprintf(stderr, "[gLExec]: Payload %d continued.\n",(int)wpid);
		/* Return controlling terminal if needed */
		if ( global_ipc_payloadstdinopts & _PAYLOAD_TAKETTY)  {
		    if (tcsetpgrp(STDIN_FILENO, wpid))  {
			global_ipc_payloadstdinopts&=~_PAYLOAD_TAKETTY;
			glexec_log(GLEXEC_LOG_WARNING,
			    "Cannot give control of tty to payload: %s\n",
			    strerror(errno));
		    } else
			glexec_log(GLEXEC_LOG_DEBUG,
			    "Payload continued, giving back tty.\n");
		}
		continue;
	    }
	}
	break;
    } while (1);

    /* Even if waitpid failed, we won't have means to figure out state of child,
     * so assume it has finished and reset timer */
    global_ipc_childstage=-3;  /* child finished */
    alarm(0);

    /* Now that payload finished, switch to root user and primary group to
     * prevent further non-root originating kills, don't make it mandatory, but
     * warn if it fails. Note: preferably don't kill secondary groups as we
     * might remove a tracking gid. */
    if (global_ipc_numsigsent==0)   /* Only necessary when ending from itself */
	glexec_continue_as_root();

    /* waitpid could give an error when signal handler has done the wait. In
     * that case, the wpid and status are in the globals */
    if (wpid < 0)   {	/* Error */
	if (wait_errno==ECHILD)	{ /* Sighandler has done our wait */
	    wpid=global_ipc_wpid;
	    status=global_ipc_status;
	    global_ipc_wpid=0;
	    global_ipc_status=0;
	} else	{
	    /* No reason to continue, because we have no status on payload */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Waitpid() failed on payload pid %d: %s\n",
		    global_ipc_childpid, strerror(errno));
	    glexec_exit(GLEXEC_UNCLASIFIED, global_ipc_childrc);
	}
    } else if (wpid==0)	{ /* Can only happen when WNOHANG is specified! */
	/* No reason to continue, because we have no status on payload */
	glexec_log(GLEXEC_LOG_ERROR,
	       "Error: this should not happen, waitpid() returned 0 while "
	       "waiting for payload pid %d.\n",
	       global_ipc_childpid);
	glexec_exit(GLEXEC_UNCLASIFIED, global_ipc_childrc);
    }

    /* Parse and log exit code of payload */
    global_ipc_childrc=glexec_parse_waitstatus(status, wpid);

    /* Look for remaining processes in pgid if we use them */
    glexec_cleanup_procesgroup(global_ipc_childpid);

    /* Take back controlling terminal when needed. */
    if ( global_ipc_payloadstdinopts & _PAYLOAD_TAKETTY)    {
	if (tcsetpgrp(STDIN_FILENO, glexec_pgid))
	    glexec_log(GLEXEC_LOG_WARNING,
		"Cannot grab control of tty from payload: %s\n",
		strerror(errno));
	else
	    glexec_log(GLEXEC_LOG_DEBUG,"Payload finished, taking back tty.\n");
    }

    /* Dump accounting info */
    glexec_accounting(1);
    
    return global_ipc_childrc;
}

/*!
 * Waits for the epilogue process to finish and correctly handles, logs etc. the
 * reason (normal exit, signal etc.).
 * \return exit code for epilogue process, following (ba)sh convention: when due
 * to a signal: 128+<signal>, in case of unsignalled but abnormal program end,
 * we return 1. See http://tldp.org/LDP/abs/html/exitcodes.html for details.
 */ 
int glexec_waitepilog(void)  {
    pid_t wpid;
    int status,wait_errno;

    /* Log at least to debug what we are doing */
    glexec_log(GLEXEC_LOG_DEBUG,"Waiting for epilogue process %d\n",
		global_ipc_epilogpid);
    
    /* Wait for pid to end, note: this is the only wait(pid), and the
     * sighandlers restart waitpid (SA_RESTART). Also wait for stopped and
     * continued processes to give feedback. */
    do	{
	wpid = waitpid(global_ipc_epilogpid, &status, WCONTINUED|WUNTRACED);
	wait_errno=errno;
	if (wpid>0) {
	    if (WIFSTOPPED(status)) {
		glexec_log(GLEXEC_LOG_WARNING,
			"Epilogue %d has suspended.\n",(int)wpid);
		continue;
	    } else if (WIFCONTINUED(status))	{
		glexec_log(GLEXEC_LOG_NOTICE,
			"Epilogue %d has continued.\n",(int)wpid);
		continue;
	    }
	}
	break;
    } while (1);

    /* Even if waitpid failed, we won't have means to figure out state of child,
     * so assume it has finished and reset timer */
    global_ipc_epilogstage=-3;  /* child finished */
    alarm(0);

    /* waitpid could give an error when signal handler has done the wait. In
     * that case, the wpid and status are in the globals */
    if (wpid < 0)   {	/* Error */
	if (wait_errno==ECHILD)	{ /* Sighandler has done our wait */
	    wpid=global_ipc_wpid;
	    status=global_ipc_status;
	    global_ipc_wpid=0;
	    global_ipc_status=0;
	} else	{
	    /* No reason to continue, because we have no status on epilogue */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error: Waitpid() failed on epilogue pid %d "
		    "(payload exitcode=%d): %s\n",
		    global_ipc_epilogpid, global_ipc_childrc, strerror(errno));
	    glexec_exit(GLEXEC_PREMATURE_EXIT, global_ipc_childrc);
	}
    } else if (wpid==0)	{ /* Can only happen when WNOHANG is specified! */
	/* No reason to continue, because we have no status on epilogue */
	glexec_log(GLEXEC_LOG_ERROR,
	       "Error: this should not happen, waitpid() returned 0 while "
	       "waiting for epilogue pid %d (payload exitcode=%d).\n",
	       global_ipc_epilogpid, global_ipc_childrc);
	glexec_exit(GLEXEC_PREMATURE_EXIT, global_ipc_childrc);
    }
    
    /* Parse and log exit code of epilogue */
    global_ipc_epilogrc=glexec_parse_waitstatus(status,wpid);

    /* Look for remaining processes in pgid if we use them */
    glexec_cleanup_procesgroup(global_ipc_epilogpid);

    return global_ipc_epilogrc;
}

/*!
 * Forks of the child process, sets the global_ipc_childpid and start the
 * accounting.
 * Failure is handled internally.
 * \return 0 child
 * \return pid parent
 */
pid_t glexec_forkchild(void)	{
    sigset_t oldset;
    pid_t glexec_pgrp, stdin_pgrp;

    /* First figure out what to do with STDIN, both for gLExec and payload */

    /* If stdin is a tty:
     * - glexec should keep it open.
     * - payload: when sharing pgids, do nothing special, when different pgids:
     *   take over the controlling tty, unless force_payload_background */
    if (isatty(STDIN_FILENO))	{
	/* gLExec keeps it open */
	global_ipc_glexecclosestdin=0;

	/* Has payload different process group? */
	if (global_ipc_usesetpgid)    {
	    /* -b cmdline or force_payload_background in glexec.conf? */
	    if (global_ipc_forcepayloadbg) {
		global_ipc_payloadstdinopts=_PAYLOAD_STDINCLOSE;
		glexec_log(GLEXEC_LOG_DEBUG,
		    "STDIN is a tty, but payload should run in background, "
		    "payload will have /dev/null\n");
	    } else { /* If glexec owns STDIN tty: set it on child */
		glexec_pgrp=getpgrp();
		stdin_pgrp=tcgetpgrp(STDIN_FILENO);
		if (stdin_pgrp==glexec_pgrp) { /* stdin is tty, but not ours */
		    global_ipc_payloadstdinopts=_PAYLOAD_TAKETTY;
		    glexec_log(GLEXEC_LOG_DEBUG,
			"Payload will take over stdin, which is a tty owned "
			"by gLExec's pgid %d\n",glexec_pgrp);
		} else {
		    global_ipc_payloadstdinopts=_PAYLOAD_STDINCLOSE;
		    glexec_log(GLEXEC_LOG_NOTICE,
			"STDIN is a tty owned by pgid %d "
			"while gLExec's pgid=%d, payload will have /dev/null\n",
			stdin_pgrp,glexec_pgrp);
		}
	    }
	} else
	    global_ipc_payloadstdinopts=0;
    } else  {
	/* No tty: child keeps open, gLExec closes */
	global_ipc_glexecclosestdin=1;
	global_ipc_payloadstdinopts=0;
    }

    /* Now do the actual fork */
    glexec_block_signals(&oldset);
    global_ipc_numsigsent=0;   /* Reset number of sent signals */
    alarm(0);	    /* Make sure no active timers */

    /* Actual fork */
    global_ipc_childstage=fork();

    if (global_ipc_childstage<0)	{ /* error */
	glexec_log(GLEXEC_LOG_ERROR ,"Fork failed: %s.\n",strerror(errno));
	glexec_exit(GLEXEC_UNCLASIFIED, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    if (global_ipc_childstage==0)	{   /* actual child */
	/* Revert umask to initial one: Note that this cannot fail */
	umask(glexec_get_caller_umask());
	/* Child should do a setpgid: do before unblocking signal */
	if (global_ipc_usesetpgid && setpgid(0,0))   {
	    glexec_log(GLEXEC_LOG_ERROR ,"Setpgid failed: %s.\n",
		    strerror(errno));
	    /* Now exit with a GLEXEC_EXIT_INTERNAL_ERROR, which will trigger a
	     * GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP, but it's the only
	     * reasonable value. */
	    exit(GLEXEC_EXIT_INTERNAL_ERROR);
	}
	/* Handle the stdin of the payload process */
	if (global_ipc_payloadstdinopts & _PAYLOAD_TAKETTY)	{
	    /* Take tty, failure not considered fatal. */
	    if (tcsetpgrp(STDIN_FILENO, getpgrp()))
		glexec_log(GLEXEC_LOG_WARNING,
		    "Payload cannot grab control of tty: %s\n",strerror(errno));
	    else
		glexec_log(GLEXEC_LOG_DEBUG,
		    "Payload has taken over tty from gLExec process.\n");
	} else	{
	    if (global_ipc_payloadstdinopts & _PAYLOAD_STDINCLOSE)
		/* Reopen /dev/null as stdin, failure not considered fatal. */
		if (freopen("/dev/null","r",stdin)==NULL)
		    glexec_log(GLEXEC_LOG_WARNING,
			"Redirecting stdin for payload failed.\n");
	}
    } else {	/* Lingering gLExec */
	/* Initialize accounting time in parent */ 
	glexec_accounting(0);
	
	/* Now also close stdin (reopen from /dev/null). Do this here, so that
	 * we can easily send a signal to just created child in case of failure
	 * */
	if (global_ipc_glexecclosestdin)	{
	    /* Reopen /dev/null as stdin, failure is considered fatal. */
	    if (freopen("/dev/null","r",stdin)==NULL)  {
		glexec_log(GLEXEC_LOG_ERROR,
		    "Redirecting stdin for lingering gLExec failed, sending "
		    "SIGTERM to payload process %s: %s\n",
		    global_ipc_usesetpgid ? "group" : "id",strerror(errno));
		/* Send signal */
		glexec_send_signal(
		    global_ipc_usesetpgid ? -global_ipc_childstage :
					    global_ipc_childstage,
		    SIGTERM, &global_ipc_childstage);
		/* exit */
		glexec_exit(GLEXEC_PREMATURE_EXIT,GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	}
    }
    global_ipc_childpid=global_ipc_childstage;
    glexec_sigprocmask(&oldset);

    return global_ipc_childpid;
}

/*!
 * Forks of the epilog process, sets the global_ipc_epilogpid and start the
 * accounting.
 * Failure is handles internally.
 * \return 0 epilog
 * \return pid parent
 */
pid_t glexec_forkepilog(void)	{
    sigset_t oldset;

    glexec_block_signals(&oldset);
    global_ipc_numsigsent=0;   /* Reset number of sent signals */
    alarm(0);	    /* Make sure no active timers */
    global_ipc_epilogstage=fork();
    if (global_ipc_epilogstage<0)	{ /* error */
	glexec_log(GLEXEC_LOG_ERROR ,"Fork failed: %s.\n",strerror(errno));
	glexec_exit(GLEXEC_UNCLASIFIED, global_ipc_childrc);
    }
    if (global_ipc_epilogstage==0)	{
	/* Child should do a setpgid */
	if (global_ipc_usesetpgid && setpgid(0,0))   {
	    glexec_log(GLEXEC_LOG_ERROR ,"setpgid failed: %s.\n",
		    strerror(errno));
	    /* No reasonable exit value other than just
	     * GLEXEC_EXIT_INTERNAL_ERROR. Note: this will trigger a
	     * 204/GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP, but that is actually
	     * correct since both child and parent should exit with an internal
	     * error. */
	    exit(GLEXEC_EXIT_INTERNAL_ERROR); 
	}
    } else {
	/* Set epilogue timeout: ONLY IN PARENT! */
	alarm(global_ipc_epilogtimeout);
    }
    global_ipc_epilogpid=global_ipc_epilogstage;
    glexec_sigprocmask(&oldset);

    return global_ipc_epilogpid;
}

/*!
 * Starts (started==0) and stops+logs (started==1) accounting info. When called
 * a second time with started==1, nothing happens.
 * \param started 0 for starting the timing records, 1 for stopping and
 * printing.
 */
void glexec_accounting(int started)	{
    static clock_t real_time_start;
    static struct tms tms_start;
    struct tms tms_stop;
    clock_t clock_ticks;
    clock_t real_time_stop;
    static int logged=0;

    if (started==0) {
	real_time_start=times(&tms_start);
    } else {
	if (logged)
	    return;

	/* Store endtime */
	real_time_stop=times(&tms_stop);

	/* Print out timing statistics */
	if ((clock_ticks = sysconf(_SC_CLK_TCK)) <= 0 ) {
	    glexec_log(GLEXEC_LOG_WARNING,
		"Clockticks/second could not be determined, "
		"using clockticks instead of seconds.\n");
	    clock_ticks = 1;
	}
	glexec_log(GLEXEC_LOG_NOTICE,
		"times: real/user/sys (%f/%f/%f)\n",
		(double)(real_time_stop - real_time_start)/clock_ticks,
		(double)(tms_stop.tms_cutime-tms_start.tms_cutime)/clock_ticks,
		(double)(tms_stop.tms_cstime-tms_start.tms_cstime)/clock_ticks);
   
	/* Prevent double logging */
	logged=1;
    }
}

